@@ -1,7 +1,7 @@
 // $Id$
 //
 // Earth System Modeling Framework
-// Copyright 2002-2014, University Corporation for Atmospheric Research, 
+// Copyright 2002-2018, University Corporation for Atmospheric Research, 
 // Massachusetts Institute of Technology, Geophysical Fluid Dynamics 
 // Laboratory, University of Michigan, National Centers for Environmental 
 // Prediction, Los Alamos National Laboratory, Argonne National Laboratory, 
@@ -32,11 +32,16 @@
 //-------------------------------------------------------------------------
 
 #include <vector>
+#include <map>
 
 #include "ESMCI_Base.h"       // Base is superclass to DELayout
 #include "ESMCI_VM.h"
 #include "ESMCI_F90Interface.h"
 
+namespace ESMCI {
+  class RouteHandle;
+}
+
 //-------------------------------------------------------------------------
 
 namespace ESMCI {
@@ -106,8 +111,14 @@ class DELayout : public ESMC_Base {    // inherits from ESMC_Base class
     
   public:
     // native constructor and destructor
-    DELayout(VM *vm=NULL):ESMC_Base(vm){} // allow specific VM instead default
-    DELayout(int baseID):ESMC_Base(baseID){}// prevent baseID counter increment
+    DELayout(VM *vm=NULL):ESMC_Base(vm){    // allow specific VM instead default
+      // initialize the name for this DELayout object in the Base class
+      ESMC_BaseSetName(NULL, "DELayout");
+    }
+    DELayout(int baseID):ESMC_Base(baseID){ // prevent baseID counter increment
+      // initialize the name for this DELayout object in the Base class
+      ESMC_BaseSetName(NULL, "DELayout");
+    }
     ~DELayout(){destruct();}
     
   private:
@@ -122,9 +133,9 @@ class DELayout : public ESMC_Base {    // inherits from ESMC_Base class
     static DELayout *create(int *petMap, int petMapCount,
       ESMC_Pin_Flag *pinFlag, VM *vm=NULL, int *rc=NULL);
     static DELayout *create(int *deCount=NULL,
-      InterfaceInt *deGrouping=NULL, ESMC_Pin_Flag *pinFlag=NULL,
-      InterfaceInt *petList=NULL, VM *vm=NULL, int *rc=NULL);
-    static int destroy(ESMCI::DELayout **layout);
+      InterArray<int> *deGrouping=NULL, ESMC_Pin_Flag *pinFlag=NULL,
+      InterArray<int> *petList=NULL, VM *vm=NULL, int *rc=NULL);
+    static int destroy(ESMCI::DELayout **layout, bool noGarbage=false);
     // get() and set()
     VM *getVM()                       const {return vm;}
     int getDeCount()                  const {return deCount;}
@@ -261,6 +272,19 @@ class XXE{
       char opInfo[16*8];      // 16 x 8-byte padding to hold stream specific
                               // members
     };
+    struct SuperVectP{
+      // Super-vectorization parameter set
+      int srcSuperVecSize_r;  // src undist
+      int srcSuperVecSize_s;  // src undist
+      int srcSuperVecSize_t;  // src undist
+      int *srcSuperVecSize_i; // src dist
+      int *srcSuperVecSize_j; // src dist
+      int dstSuperVecSize_r;  // dst undist
+      int dstSuperVecSize_s;  // dst undist
+      int dstSuperVecSize_t;  // dst undist
+      int *dstSuperVecSize_i; // dst dist
+      int *dstSuperVecSize_j; // dst dist
+    };
     
     // special predefined filter bits
     static int const filterBitRegionTotalZero   = 0x1;  // total dst zero'ing
@@ -278,33 +302,42 @@ class XXE{
       // information is also provided to support buffer resizing during exec(),
       // when the actual execution time vectorLength is known.
       char *buffer;                 // buffer
-      int size;                     // size of buffer in byte
+      unsigned long size;           // size of buffer in byte
       int vectorLengthMultiplier;   // multiplier that allows to determine size
                                     // requirement depending on vectorLength
                                     // during exec()
       
-      BufferInfo(char *buffer_, int size_, int vectorLengthMultiplier_){
+      BufferInfo(char *buffer_, unsigned long size_, 
+        int vectorLengthMultiplier_){
         // constructor
         buffer = buffer_;
         size = size_;
         vectorLengthMultiplier = vectorLengthMultiplier_;
       }
     };
-        
+    
   public:
     VM *vm;
-    StreamElement *stream;          // actual stream containing XXE elements
+    // OPSTREAM
+    StreamElement *opstream;        // actual stream containing XXE operations
     int count;                      // number of elements in the stream
-    char **storage;                 // list of (char *) entries to allocations
+    // DATA
+    std::map<void *, unsigned long>
+      dataMap;                      // map object to hold (pointer,size) pairs
+    char **dataList;                // list of (char *) entries to allocations
                                     // for which this XXE object is responsible
-    int storageCount;               // number of elements in storage
+    int dataCount;                  // number of elements in dataList
+    // COMMHANDLES
     VMK::commhandle ***commhandle;  // list of (commhandle **) entries for 
                                     // which this XXE object is responsible
     int commhandleCount;            // number of elements in commhandle
+    // SUBXXES
     XXE **xxeSubList;               // list of (XXE *) entries for which this
                                     // XXE object is responsible
     int xxeSubCount;                // number of elements in xxeSubList
+    // TYPEKINDS
     ESMC_TypeKind_Flag typekind[10];     // place the XXE can store TypeKind info
+    // BUFFERS
     std::vector<BufferInfo *>bufferInfoList; // vector of (BufferInfo *) entries
       // The bufferInfoList provides an extra level of indirection to XXE
       // managed communication buffers, and associated size information.
@@ -316,23 +349,26 @@ class XXE{
       // updates during exec() through the bufferInfoList indirection, without
       // the need for XXE stream rewrite (which would be far too expensive to
       // do during exec())!
+    // MISC
     int lastFilterBitField;         // filterBitField during last exec() call
+    bool superVectorOkay;           // flag to indicate that super-vector okay
   private:
     int max;                        // maximum number of elements in stream
-    int storageMaxCount;            // maximum number of elements in storage
+    int dataMaxCount;               // maximum number of elements in data
     int commhandleMaxCount;         // maximum number of elements in commhandle
     int xxeSubMaxCount;             // maximum number of elements in xxeSubList
+    RouteHandle *rh;                // associated RouteHandle
     
   public:
-    XXE(VM *vmArg, int maxArg=1000, int storageMaxCountArg=1000,
+    XXE(VM *vmArg, int maxArg=1000, int dataMaxCountArg=1000,
       int commhandleMaxCountArg=1000, int xxeSubMaxCountArg=1000){
-      // constructor
+      // constructor...
       vm = vmArg;
-      // -> set up internal stream and bookkeeping members
-      stream = new StreamElement[maxArg]; count = 0; max = maxArg;
-      storage = new char*[storageMaxCountArg];
-      storageCount  = 0;
-      storageMaxCount = storageMaxCountArg;
+      // -> set up internal opstream and bookkeeping members
+      opstream = new StreamElement[maxArg]; count = 0; max = maxArg;
+      dataList = new char*[dataMaxCountArg];
+      dataCount  = 0;
+      dataMaxCount = dataMaxCountArg;
       commhandle = new VMK::commhandle**[commhandleMaxCountArg];
       commhandleCount  = 0;
       commhandleMaxCount = commhandleMaxCountArg;
@@ -341,66 +377,42 @@ class XXE{
       xxeSubMaxCount = xxeSubMaxCountArg;
       bufferInfoList.reserve(1000);  // initial preparation
       lastFilterBitField = 0x0;
+      superVectorOkay = true;
+      rh = NULL;
     }
-    ~XXE(){
-      // destructor
-      // -> clean-up all allocations for which this XXE object is responsible:
-      // stream of XXE elements
-      delete [] stream;
-      // memory allocations held in storage
-      for (int i=0; i<storageCount; i++)
-        delete [] storage[i];
-      delete [] storage;
-      // CommHandles held in commhandle
-      for (int i=0; i<commhandleCount; i++){
-        delete *commhandle[i];
-        delete commhandle[i];
-      }
-      delete [] commhandle;
-      // XXE sub objects held in xxeSubList
-      for (int i=0; i<xxeSubCount; i++)
-        delete xxeSubList[i];
-      delete [] xxeSubList;
-      // BufferInfo objects held in bufferInfoList
-      for (int i=0; i<bufferInfoList.size(); i++)
-        delete bufferInfoList[i];
-      bufferInfoList.clear();
-    }
-    void clearReset(int countArg, int storageCountArg=-1, 
+    XXE(std::stringstream &streami,
+      std::vector<int> *originToTargetMap=NULL,
+      std::map<void *, void *> *bufferOldNewMap=NULL,
+      std::map<void *, void *> *dataOldNewMap=NULL);
+    ~XXE();      // destructor
+    void clearReset(int countArg, int dataCountArg=-1, 
       int commhandleCountArg=-1, int xxeSubCountArg=-1, 
-      int bufferInfoListArg=-1){
-      // reset the stream back to a specified position, and clear all
-      // bookkeeping elements above specified positions
-      count = countArg; // reset
-      if (storageCountArg>0){
-        for (int i=storageCountArg; i<storageCount; i++)
-          delete [] storage[i];
-        storageCount = storageCountArg; // reset
-      }
-      if (commhandleCountArg>0){
-        for (int i=commhandleCountArg; i<commhandleCount; i++){
-          delete *commhandle[i];
-          delete commhandle[i];
+      int bufferInfoListArg=-1);
+    void streamify(std::stringstream &streami);
+    bool getNextSubSuperVectorOkay(int *k){
+      // search for the next xxeSub element in the opstream, starting at 
+      // index k. when found return the element's superVectorOkay setting,
+      // and also update k to point to the next stream element for continued
+      // search
+      XxeSubInfo *xxeSubInfo;
+      for (int i=*k; i<count; i++){
+        if (opstream[i].opId==xxeSub){
+          *k = i+1;   // index where to start next time
+          xxeSubInfo = (XxeSubInfo *)&(opstream[i]);
+          if (xxeSubInfo->xxe)
+            return xxeSubInfo->xxe->superVectorOkay;
+          else
+            return false; // default
         }
-        commhandleCount = commhandleCountArg; // reset
-      }
-      if (xxeSubCountArg>0){
-        for (int i=xxeSubCountArg; i<xxeSubCount; i++)
-          delete xxeSubList[i];
-        xxeSubCount = xxeSubCountArg; // reset
-      }
-      if (bufferInfoListArg>0){
-        std::vector<BufferInfo *>::iterator first =
-          bufferInfoList.begin() + bufferInfoListArg;
-        std::vector<BufferInfo *>::iterator last = bufferInfoList.end();
-        for (std::vector<BufferInfo *>::iterator bi=first; bi!=last; ++bi)
-          delete *bi;
-        bufferInfoList.erase(first, last);
       }
+      // if no subXXE found then return false
+      return false;
     }
+    
     int exec(int rraCount=0, char **rraList=NULL, int *vectorLength=NULL,
       int filterBitField=0x0, bool *finished=NULL, bool *cancelled=NULL, 
-      double *dTime=NULL, int indexStart=-1, int indexStop=-1);
+      double *dTime=NULL, int indexStart=-1, int indexStop=-1,
+      int *srcLocalDeCount=NULL, SuperVectP *superVectP=NULL);
     int print(FILE *fp, int rraCount=0, char **rraList=NULL,
       int filterBitField=0x0, int indexStart=-1, int indexStop=-1);
     int printProfile(FILE *fp);
@@ -409,19 +421,20 @@ class XXE{
     int optimizeElement(int index);
     
     int growStream(int increase);
-    int growStorage(int increase);
+    int growDataList(int increase);
     int growCommhandle(int increase);
     int growXxeSub(int increase);
     
     int incCount();
-    int incStorageCount();
+    int incDataCount();
     int incCommhandleCount();
     int incXxeSubCount();
     
-    int storeStorage(char *storage);
+    int storeData(char *data, unsigned long size);
     int storeCommhandle(VMK::commhandle **commhandle);
     int storeXxeSub(XXE *xxeSub);
-    int storeBufferInfo(char *buffer, int size, int vectorLengthMultiplier);
+    int storeBufferInfo(char *buffer, unsigned long size,
+      int vectorLengthMultiplier);
     char *getBufferInfoPtr(){
       size_t i=bufferInfoList.size();
       if (i>0)
@@ -443,7 +456,8 @@ class XXE{
       int srcSize, int dstSize, int srcPet, int dstPet, int srcTag, int dstTag, 
       bool vectorFlag=false, bool srcIndirectionFlag=false,
       bool dstIndirectionFlag=false);
-    int appendSendRRARecv(int predicateBitField, int rraOffset, void *dstBuffer,       int srcSize, int dstSize, int srcPet, int dstPet, int rraIndex,
+    int appendSendRRARecv(int predicateBitField, int rraOffset, void *dstBuffer,
+      int srcSize, int dstSize, int srcPet, int dstPet, int rraIndex,
       int srcTag, int dstTag, bool vectorFlag=false,
       bool dstIndirectionFlag=false);
     int appendRecvnb(int predicateBitField, void *buffer, int size, int srcPet,
@@ -499,45 +513,16 @@ class XXE{
     int appendProfileMessage(int predicateBitField, char *messageString);
     int appendMessage(int predicateBitField, char *messageString);
     
-  private:
-    template<typename T, typename U, typename V>
-    static void psv(T *element, TKId elementTK, U *factorList, TKId factorTK,
-      V *valueList, TKId valueTK, int factorCount, int resolved);
-    template<typename T, typename U, typename V>
-    static void pss(T *element, TKId elementTK, U *factor, TKId factorTK,
-      V *value, TKId valueTK, int resolved);
-    template<typename T, typename V>
-    static void sssDstRra(T *rraBase, TKId elementTK, int *rraOffsetList,
-      V *valueBase, int *valueOffsetList, TKId valueTK, int termCount,
-      int vectorLength, int resolved);
-    template<typename T, typename V>
-    static void ssslDstRra(T **rraBaseList, int *rraIndexList, TKId elementTK,
-      int *rraOffsetList, V **valueBaseList,
-      int *valueOffsetList, int *baseListIndexList,
-      TKId valueTK, int termCount, int vectorLength, int resolved);
-    template<typename T, typename U, typename V>
-    static void psssDstRra(T *rraBase, TKId elementTK, int *rraOffsetList,
-      U **factorList, TKId factorTK, V *valueBase, int *valueOffsetList,
-      TKId valueTK, int termCount, int vectorLength, int resolved);
-    template<typename T, typename U, typename V>
-    static void pssslDstRra(T **rraBaseList, int *rraIndexList, TKId elementTK,
-      int *rraOffsetList, U **factorList, TKId factorTK, V **valueBaseList,
-      int *valueOffsetList, int *baseListIndexList,
-      TKId valueTK, int termCount, int vectorLength, int resolved);
-    template<typename T, typename U, typename V>
-    static void psssSrcRra(T *rraBase, TKId valueTK, int *rraOffsetList,
-      U **factorList, TKId factorTK, V *elementBase, int *elementOffsetList,
-      TKId elementTK, int termCount, int vectorLength, int resolved);
-    template<typename T, typename U, typename V>
-    static void pssscRra(T *rraBase, TKId elementTK, int *rraOffsetList,
-      U **factorList, TKId factorTK, V *valueList, TKId valueTK,
-      int termCount, int vectorLength, int resolved);
-
+    void setRouteHandle(RouteHandle *routehandle){
+      rh = routehandle;
+    }
+    RouteHandle *getRouteHandle(){return rh;}
+        
   public:
       
-    // specific stream element types, used to interpret the elements in stream
+    // opstream element types, used to interpret the elements in opstream
     
-    // Common stream element members and their meaning:
+    // Common opstream element members and their meaning:
     //  opId                - operation id according to "enum OpId"
     //  predicateBitField   - predicate bit field to control conditional exec
     //  vectorFlag          - true:  scale with vectorLength during exec()
@@ -550,12 +535,12 @@ class XXE{
       int predicateBitField;
       bool activeFlag;
       bool cancelledFlag;
-      void *buffer;
-      int size;
       int dstPet;
-      int tag;
       bool vectorFlag;
       bool indirectionFlag;
+      void *buffer;
+      int size;
+      int tag;
     }SendInfo;
 
     typedef struct{
@@ -563,12 +548,12 @@ class XXE{
       int predicateBitField;
       bool activeFlag;
       bool cancelledFlag;
-      void *buffer;
-      int size;
       int srcPet;
-      int tag;
       bool vectorFlag;
       bool indirectionFlag;
+      void *buffer;
+      int size;
+      int tag;
     }RecvInfo;
 
     typedef struct{
@@ -576,12 +561,12 @@ class XXE{
       int predicateBitField;
       bool activeFlag;
       bool cancelledFlag;
+      int dstPet;
+      bool vectorFlag;
       int rraOffset;
       int size;
-      int dstPet;
       int rraIndex;
       int tag;
-      bool vectorFlag;
     }SendRRAInfo;
 
     typedef struct{
@@ -589,12 +574,12 @@ class XXE{
       int predicateBitField;
       bool activeFlag;
       bool cancelledFlag;
+      int srcPet;
+      bool vectorFlag;
       int rraOffset;
       int size;
-      int srcPet;
       int rraIndex;
       int tag;
-      bool vectorFlag;
     }RecvRRAInfo;
 
     typedef struct{
@@ -602,17 +587,17 @@ class XXE{
       int predicateBitField;
       bool activeFlag;
       bool cancelledFlag;
+      int srcPet;
+      int dstPet;
+      bool vectorFlag;
+      bool srcIndirectionFlag;
+      bool dstIndirectionFlag;
       void *srcBuffer;
       void *dstBuffer;
       int srcSize;
       int dstSize;
-      int srcPet;
-      int dstPet;
       int srcTag;
       int dstTag;
-      bool vectorFlag;
-      bool srcIndirectionFlag;
-      bool dstIndirectionFlag;
     }SendRecvInfo;
 
     typedef struct{
@@ -620,17 +605,17 @@ class XXE{
       int predicateBitField;
       bool activeFlag;
       bool cancelledFlag;
+      int srcPet;
+      int dstPet;
+      bool vectorFlag;
+      bool dstIndirectionFlag;
       int rraOffset;
       void *dstBuffer;
       int srcSize;
       int dstSize;
-      int srcPet;
-      int dstPet;
       int rraIndex;
       int srcTag;
       int dstTag;
-      bool vectorFlag;
-      bool dstIndirectionFlag;
     }SendRRARecvInfo;
 
     typedef struct{
@@ -639,12 +624,12 @@ class XXE{
       VMK::commhandle **commhandle;
       bool activeFlag;
       bool cancelledFlag;
-      void *buffer;
-      int size;
       int dstPet;
-      int tag;
       bool vectorFlag;
       bool indirectionFlag;
+      void *buffer;
+      int size;
+      int tag;
     }SendnbInfo;
 
     typedef struct{
@@ -653,12 +638,12 @@ class XXE{
       VMK::commhandle **commhandle;
       bool activeFlag;
       bool cancelledFlag;
-      void *buffer;
-      int size;
       int srcPet;
-      int tag;
       bool vectorFlag;
       bool indirectionFlag;
+      void *buffer;
+      int size;
+      int tag;
     }RecvnbInfo;
 
     typedef struct{
@@ -667,12 +652,12 @@ class XXE{
       VMK::commhandle **commhandle;
       bool activeFlag;
       bool cancelledFlag;
+      int dstPet;
+      bool vectorFlag;
       int rraOffset;
       int size;
-      int dstPet;
       int rraIndex;
       int tag;
-      bool vectorFlag;
     }SendnbRRAInfo;
 
     typedef struct{
@@ -681,12 +666,12 @@ class XXE{
       VMK::commhandle **commhandle;
       bool activeFlag;
       bool cancelledFlag;
+      int srcPet;
+      bool vectorFlag;
       int rraOffset;
       int size;
-      int srcPet;
       int rraIndex;
       int tag;
-      bool vectorFlag;
     }RecvnbRRAInfo;
 
     typedef struct{
@@ -814,7 +799,7 @@ class XXE{
       TKId factorTK;
       TKId valueTK;
       int *rraOffsetList;
-      void **factorList;
+      void *factorList;
       void *valueBase;
       int rraIndex;
       int termCount;
@@ -830,7 +815,7 @@ class XXE{
       TKId factorTK;
       TKId valueTK;
       int *rraOffsetList;
-      void **factorList;
+      void *factorList;
       void **valueBaseList;
       void **valueBaseListResolve;
       int valueBaseListSize;
@@ -849,7 +834,7 @@ class XXE{
       TKId factorTK;
       TKId valueTK;
       int *rraOffsetList;
-      void **factorList;
+      void *factorList;
       void *elementBase;
       int rraIndex;
       int termCount;
@@ -865,7 +850,7 @@ class XXE{
       TKId factorTK;
       TKId valueTK;
       int *rraOffsetList;
-      void **factorList;
+      void *factorList;
       void *valueList;
       int rraIndex;
       int termCount;
@@ -894,10 +879,10 @@ class XXE{
     typedef struct{
       OpId opId;
       int predicateBitField;
-      void *buffer;
-      int byteCount;
       bool vectorFlag;
       bool indirectionFlag;
+      void *buffer;
+      int byteCount;
     }ZeroMemsetInfo;
 
     typedef struct{
@@ -938,7 +923,7 @@ class XXE{
       bool indirectionFlag;
     }MemGatherSrcRRAInfo;
     
-    // --- sub-streams
+    // --- sub-opstreams
     
     typedef struct{
       OpId opId;
@@ -987,7 +972,7 @@ class XXE{
       char *messageString;
     }ProfileMessageInfo;
     
-    // --- meta Info structs (i.e. don't correspond to OpIds)
+    // --- meta Info structs (i.e. don't correspond to specific OpIds)
 
     typedef struct{
       OpId opId;
@@ -995,7 +980,167 @@ class XXE{
       VMK::commhandle **commhandle;
       bool activeFlag;
       bool cancelledFlag;
+      int pet;
     }CommhandleInfo;
+    
+    typedef struct{
+      OpId opId;
+      int predicateBitField;
+      bool activeFlag;
+      bool cancelledFlag;
+      int pet;
+      bool vectorFlag;
+      bool indirectionFlag;
+      void *buffer;
+      int size;
+    }BuffInfo;
+    
+    typedef struct{
+      OpId opId;
+      int predicateBitField;
+      VMK::commhandle **commhandle;
+      bool activeFlag;
+      bool cancelledFlag;
+      int pet;
+      bool vectorFlag;
+      bool indirectionFlag;
+      void *buffer;
+      int size;
+    }BuffnbInfo;
+
+    typedef struct{
+      OpId opId;
+      int predicateBitField;
+      XXE *xxe;
+    }SingleSubInfo;
+    
+    typedef struct{
+      OpId opId;
+      int predicateBitField;
+      int count;
+      XXE **xxe;
+    }MultiSubInfo;
+    
+  private:
+    template<typename T>
+    inline static void exec_memGatherSrcRRA(
+      MemGatherSrcRRAInfo *xxeMemGatherSrcRRAInfo, int vectorL, char **rraList);
+    template<typename T>
+    inline static void exec_memGatherSrcRRASuper(
+      MemGatherSrcRRAInfo *xxeMemGatherSrcRRAInfo, int vectorL, char **rraList,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j);
+    template<typename T>
+    inline static void exec_zeroSuperScalarRRA(
+      ZeroSuperScalarRRAInfo *xxeZeroSuperScalarRRAInfo, int vectorL, 
+      char **rraList);
+    template<typename T>
+    inline static void exec_zeroSuperScalarRRASuper(
+      ZeroSuperScalarRRAInfo *xxeZeroSuperScalarRRAInfo, int vectorL, 
+      char **rraList, int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j);
+    template<typename T, typename U, typename V>
+    static void psv(T *element, TKId elementTK, U *factorList, TKId factorTK,
+      V *valueList, TKId valueTK, int factorCount, int resolved);
+    template<typename T, typename U, typename V>
+    static void pss(T *element, TKId elementTK, U *factor, TKId factorTK,
+      V *value, TKId valueTK, int resolved);
+    template<typename T, typename V>
+    static void sssDstRra(T *rraBase, TKId elementTK, int *rraOffsetList,
+      V *valueBase, int *valueOffsetList, TKId valueTK, int termCount,
+      int vectorL, int resolved, int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j,
+      bool superVector);
+    template<typename T, typename V>
+    static void exec_sssDstRra(T *rraBase, int *rraOffsetList, V *valueBase,
+      int *valueOffsetList, int termCount, int vectorL);
+    template<typename T, typename V>
+    static void exec_sssDstRraSuper(T *rraBase, int *rraOffsetList,
+      V *valueBase, int *valueOffsetList, int termCount, int vectorL,
+      int localDeIndexOff, int size_r, int size_s, int size_t, int *size_i,
+      int *size_j);
+    template<typename T, typename V>
+    static void ssslDstRra(T **rraBaseList, int *rraIndexList, TKId elementTK,
+      int *rraOffsetList, V **valueBaseList,
+      int *valueOffsetList, int *baseListIndexList,
+      TKId valueTK, int termCount, int vectorL, int resolved,
+      int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j,
+      bool superVector);
+    template<typename T, typename V>
+    static void exec_ssslDstRra(T **rraBaseList, int *rraIndexList,
+      int *rraOffsetList, V **valueBaseList, int *valueOffsetList,
+      int *baseListIndexList, int termCount, int vectorL);
+    template<typename T, typename V>
+    static void exec_ssslDstRraSuper(T **rraBaseList, int *rraIndexList,
+      int *rraOffsetList, V **valueBaseList, int *valueOffsetList,
+      int *baseListIndexList, int termCount, int vectorL, int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j);
+    template<typename T, typename U, typename V>
+    static void psssDstRra(T *rraBase, TKId elementTK, int *rraOffsetList,
+      U *factorList, TKId factorTK, V *valueBase, int *valueOffsetList,
+      TKId valueTK, int termCount, int vectorL, int resolved,
+      int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j,
+      bool superVector);
+    template<typename T, typename U, typename V>
+    static void exec_psssDstRra(T *rraBase, int *rraOffsetList, U *factorList,
+      V *valueBase, int *valueOffsetList, int termCount, int vectorL);
+    template<typename T, typename U, typename V>
+    static void exec_psssDstRraSuper(T *rraBase, int *rraOffsetList,
+      U *factorList, V *valueBase, int *valueOffsetList, int termCount,
+      int vectorL, int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j);
+    template<typename T, typename U, typename V>
+    static void pssslDstRra(T **rraBaseList, int *rraIndexList, TKId elementTK,
+      int *rraOffsetList, U *factorList, TKId factorTK, V **valueBaseList,
+      int *valueOffsetList, int *baseListIndexList,
+      TKId valueTK, int termCount, int vectorL, int resolved, 
+      int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j, 
+      bool superVector, RouteHandle *rh);
+    template<typename T, typename U, typename V>
+    static void exec_pssslDstRra(T **rraBaseList, int *rraIndexList, 
+      int *rraOffsetList, U *factorList, V **valueBaseList,
+      int *valueOffsetList, int *baseListIndexList,
+      int termCount, int vectorL);
+    template<typename T, typename U, typename V>
+    static void exec_pssslDstRraDynMask(T **rraBaseList, int *rraIndexList, 
+      int *rraOffsetList, U *factorList, V **valueBaseList,
+      int *valueOffsetList, int *baseListIndexList,
+      int termCount, int vectorL, RouteHandle *rh);
+    template<typename T, typename U, typename V>
+    static void exec_pssslDstRraSuper(T **rraBaseList, int *rraIndexList, 
+      int *rraOffsetList, U *factorList, V **valueBaseList,
+      int *valueOffsetList, int *baseListIndexList,
+      int termCount, int vectorL, int localDeIndexOff,
+      int size_r, int size_s, int size_t, int *size_i, int *size_j);
+    template<typename T, typename U, typename V>
+    static void psssSrcRra(T *rraBase, TKId valueTK, int *rraOffsetList,
+      U *factorList, TKId factorTK, V *elementBase, int *elementOffsetList,
+      TKId elementTK, int termCount, int vectorL, int resolved, 
+      int localDeIndexOff, int size_r, int size_s, int size_t, int *size_i, 
+      int *size_j, bool superVector);
+    template<typename T, typename U, typename V>
+    static void exec_psssSrcRra(T *rraBase, int *rraOffsetList, U *factorList,
+      V *elementBase, int *elementOffsetList, int termCount, int vectorL);
+    template<typename T, typename U, typename V>
+    static void exec_psssSrcRraSuper(T *rraBase, int *rraOffsetList,
+      U *factorList, V *elementBase, int *elementOffsetList, int termCount,
+      int vectorL, int localDeIndexOff, int size_r, int size_s, int size_t,
+      int *size_i, int *size_j, bool superVector);
+    template<typename T, typename U, typename V>
+    static void pssscRra(T *rraBase, TKId elementTK, int *rraOffsetList,
+      U *factorList, TKId factorTK, V *valueList, TKId valueTK,
+      int termCount, int vectorL, int resolved);
+
+    template<typename T, typename U, typename V> struct DynMaskElement{
+      T *element;
+      std::vector<U*> factors;
+      std::vector<V*> values;
+    };
+    template<typename T, typename U, typename V>
+    static void dynMaskHandler(std::vector<DynMaskElement<T,U,V> > &dynMaskList,
+      RouteHandle *rh, int vectorL);
 
 };  // class XXE
 
