@@ -1,6 +1,6 @@
 // $Id$
 // Earth System Modeling Framework
-// Copyright 2002-2014, University Corporation for Atmospheric Research, 
+// Copyright 2002-2018, University Corporation for Atmospheric Research, 
 // Massachusetts Institute of Technology, Geophysical Fluid Dynamics 
 // Laboratory, University of Michigan, National Centers for Environmental 
 // Prediction, Los Alamos National Laboratory, Argonne National Laboratory, 
@@ -14,6 +14,11 @@
 
 namespace ESMCI {
 
+// Regrid Status return
+#define ESMCI_REGRIDSTATUS_OUTSIDE_SRC
+#define ESMCI_REGRIDSTATUS_SRC_MASKED
+#define ESMCI_REGRIDSTATUS_DST_MASKED
+
 
 // What to do if dest points aren't found
 #define ESMCI_UNMAPPEDACTION_ERROR  0
