@@ -116,6 +116,7 @@ extern int MPIUNI_Memcpy(void*,const void*,int);
 #define MPI_INT             sizeof(int)
 #define MPI_UNSIGNED        sizeof(unsigned int)
 #define MPI_LONG            sizeof(long)
+#define MPI_LONG_LONG_INT   sizeof(long long int)
 #define MPI_SHORT           sizeof(short)
 #define MPI_UB              sizeof(long)
 #define MPI_FLOAT_INT       (sizeof(float)+sizeof(int))
@@ -130,6 +131,9 @@ extern int MPIUNI_Memcpy(void*,const void*,int);
 #define MPI_MAX_ERROR_STRING 512
 typedef long MPI_Aint;
 
+/* for moab to build in mpiuni mode */
+#define MPI_Fint  int
+
 /* Collective operators.  Same values used in mpif.h */
 typedef int MPI_Op;
 
